import numpy as np
import cv2
from ..artist import Artist, constant
from ..primitives import _DEFAULT_COLOR, _DEFAULT_SECONDARY_COLOR
from . import RGBa_interpolation, LABa_interpolation


class linear_gradient(Artist):
    color0 = constant(_DEFAULT_COLOR)
    color1 = constant(_DEFAULT_COLOR)
    theta = constant(np.pi / 4)
    interpolation = constant("LAB")

    args = ("color0", "color1", "theta", "interpolation")

    def __call__(self, cvs, t=0.0, mask=None):
        """
        Assume the mask is of type Canvas. 
        Interpolation can be LAB or RGB.
        Theta controls the angle along the shape.

        If mask is None, return True to show that a gradient exists
        """
        if mask is None:
            return True

        mask_idx = mask.alpha > 0

        # If the mask_idx is 0, image is off the canvas
        if not mask_idx.sum():
            return True

        theta = self.theta(t)

        A = np.array([np.cos(theta), np.sin(theta)])

        # Project each masked grid point onto the angle mapped by theta
        xg, yg = cvs.grid_points()
        B = np.vstack([xg[mask_idx], yg[mask_idx]])

        pro = A.dot(B)
        pro -= pro.min()
        pro /= pro.max()

        c0 = cvs.transform_color(self.color0(t))
        c1 = cvs.transform_color(self.color1(t))

        # Smooth the image based off the alpha from the mask image
        alpha = (mask.alpha / 255.0)[mask_idx]

        imode = self.interpolation(t)
        if imode == "LAB":
            C = LABa_interpolation(pro, c0, c1, alpha)
        elif imode == "RGB":
            # RGBA interpolation (not great!)
            C = RGBa_interpolation(pro, c0, c1, alpha)
        else:
            raise KeyError(f"Unknown interpolation {imode}")

        # Blend the new shape in
        rhs = cvs.copy()
        rhs._img[mask_idx] = C
        cvs.blend(rhs)



import numpy as np
import collections
from scipy.misc import comb

"""
Adapeted from https://github.com/reptillicus/Bezier
"""


def bernstein_poly(i, n, t):
    """
     The Bernstein polynomial of n, i as a function of t
    """

    return comb(n, i) * (t ** (n - i)) * (1 - t) ** i


class bezierMotionCurve:
    """
    Bezier curve with endpoints set at (0,0) and (1,1)
    """

    def __init__(self, x1, y1, x2, y2):
        points = np.array([(0, 0), (x1, y1), (x2, y2), (1, 1)])
        self.xPoints = np.array([p[0] for p in points])
        self.yPoints = np.array([p[1] for p in points])
        self.nPoints = 4

    def __call__(self, t):
        if not isinstance(t, collections.Iterable):
            t = np.array([t])

        polynomial_array = np.array(
            [
                bernstein_poly(i, self.nPoints - 1, t)
                for i in range(0, self.nPoints)
            ]
        )

        xvals = np.dot(self.xPoints, polynomial_array)
        yvals = np.dot(self.yPoints, polynomial_array)

        return xvals, yvals



from scipy.ndimage.interpolation import map_coordinates
from scipy.ndimage.filters import gaussian_filter
import numpy as np

import cv2
from ..artist import Artist, constant


class ElasticTransform(Artist):
    @staticmethod
    def transform(cvs, dy, dx, coords, mode):

        xg, yg, zg = coords

        indices = (
            np.reshape(yg + dy, (-1, 1)),
            np.reshape(xg + dx, (-1, 1)),
            np.reshape(zg, (-1, 1)),
        )

        distored_image = map_coordinates(cvs.img, indices, order=2, mode=mode)

        cvs._img = distored_image.reshape(cvs.shape)


class pull(ElasticTransform):
    x = constant(0.0)
    y = constant(0.0)
    alpha = constant(1.0)
    mode = constant("nearest")
    args = ("x", "y", "sigma", "alpha", "mode")

    def draw(self, cvs, t=0.0):
        # https://gist.github.com/erniejunior/601cdf56d2b424757de5

        x = float(cvs.transform_x(self.x(t)))
        y = float(cvs.transform_y(self.y(t)))
        alpha = cvs.transform_length(self.alpha(t), is_discrete=False)

        coords = cvs.grid_coordinates()

        theta = np.arctan2(coords[1] - y, coords[0] - x)
        dx = alpha * np.cos(theta)
        dy = alpha * np.sin(theta)

        self.transform(cvs, dy, dx, coords, self.mode(t))


class distort(ElasticTransform):
    sigma = constant(0.1)
    alpha = constant(10.0)
    mode = constant("nearest")
    seed = constant(None)

    args = ("sigma", "alpha", "mode", "seed")

    def draw(self, cvs, t=0.0):
        # https://gist.github.com/erniejunior/601cdf56d2b424757de5

        sigma = cvs.transform_length(self.sigma(t))
        alpha = cvs.transform_length(self.alpha(t), is_discrete=False)
        mode = self.mode(t)

        shape = cvs.shape
        random_state = np.random.RandomState(self.seed(t))

        coords = cvs.grid_coordinates()

        dx = random_state.rand(*shape) * 2 - 1
        dy = random_state.rand(*shape) * 2 - 1

        dx = gaussian_filter(dx, sigma, mode=mode)
        dy = gaussian_filter(dy, sigma, mode=mode)

        self.transform(cvs, dy * alpha, dx * alpha, coords, self.mode(t))


class motion_lines(ElasticTransform):
    alpha = constant(0.15)
    theta = constant(0.0)
    mode = constant("constant")
    args = ("alpha", "mode")

    def draw(self, cvs, t=0.0):
        alpha = cvs.transform_length(self.alpha(t), is_discrete=False)

        theta = self.theta(t)
        mode = self.mode(t)

        coords = cvs.grid_coordinates()

        y = cvs.inverse_transform_y(coords[1].astype(float))
        x = cvs.inverse_transform_y(coords[0].astype(float))

        dx = np.cos(theta) * np.abs(x)
        dy = np.sin(theta) * np.abs(y)

        self.transform(cvs, alpha * dy, alpha * dx, coords, self.mode(t))



"""
Version: 0.1.2: Fixed MANIFEST! Missing assets
Version: 0.1.1: Refactored .filter to .filters to prevent namespace collision
Version: 0.1.0: First release! Added subpixel resolution to prior work.
"""

__version__ = "0.1.2"



from .colors import NamedColors
from .palettes import ColorLoversPalette

from .interpolation import RGBa_interpolation
from .interpolation import LABa_interpolation

from .gradient import linear_gradient

matplotlib_colors = NamedColors()



import numpy as np
import cv2
from ..artist import Artist, constant
import warnings


class gaussian_blur(Artist):
    blur_x = constant(0.1)
    blur_y = constant(0.1)

    args = ("blur_x", "blur_y")

    def draw(self, cvs, t=0.0):
        bx = cvs.transform_kernel_length(self.blur_x(t))
        by = cvs.transform_kernel_length(self.blur_y(t))
        kernel = (bx, by)

        cvs._img = cv2.GaussianBlur(cvs.img, kernel, 0)



"""
Pseudo-instagram filters (weights derived from samples).
"""

from ..artist import Artist, constant
import numpy as np
import cv2
import glob
import os

_script_path = os.path.dirname(os.path.abspath(__file__))
_model_path = os.path.join(_script_path, "insta", "models")

known_models = set(
    [
        os.path.basename(name).split(".")[0]
        for name in glob.glob(os.path.join(_model_path, "*.npz"))
    ]
)


class instafilter(Artist):

    weight = constant(1.0)
    args = ("weight",)

    def __init__(self, name, **kwargs):
        super().__init__(**kwargs)

        if name not in known_models:
            msg = f"Model {name} not in {known_models}"
            raise KeyError(msg)

        f_model = os.path.join(_model_path, f"{name}.npz")
        assert os.path.exists(f_model)

        obj = np.load(f_model)
        self.weights = obj["W"]
        self.bias = obj["b"]

    @staticmethod
    def _saturation_and_lightness(xBGR):
        # Returns the max BGR color and the largest delta in BGR space.
        mx, mn = xBGR.max(axis=1), xBGR.min(axis=1)
        avg = xBGR.mean(axis=1)
        return np.array([avg, mx - mn]).T

    @staticmethod
    def _scale(img):
        # Scales the colors to a flat (h*w, channel) array \in [0, 1]
        h, w, channels = img.shape
        return img.reshape(h * w, channels).astype(np.float32) / 255

    def draw(self, cvs, t=0.0):

        weight = self.weight(t)
        if weight <= 0:
            # With zero weight, skip the filter
            return None

        height, width, channels = cvs.shape

        img = cvs.img[:, :, :3]
        alpha = None

        if channels != 3:
            alpha = cvs.img[:, :, 3:]

        # Model expect BGR, canvas uses RGB
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        xBGR = self._scale(img)
        xLS = self._saturation_and_lightness(xBGR)

        y = np.hstack([xBGR, xLS])

        # Apply MLP layers
        for w, b in zip(self.weights[:-1], self.bias[:-1]):
            y = np.tanh(y.dot(w) + b)

        # The last layer has no activation
        y = y.dot(self.weights[-1]) + self.bias[-1]

        yp = np.clip(y * 255, 0, 255).astype(np.uint8)
        imgBGR = yp[:, :3].reshape(height, width, 3)

        # Convert back to RGB colorspace
        img2 = cv2.cvtColor(imgBGR, cv2.COLOR_BGR2RGB)

        if alpha is not None:
            img2 = np.dstack([img2, alpha])

        # With weight > 1 apply the whole filter
        if weight >= 1:
            cvs._img = img2
        # Otherwise blend
        cvs._img = cv2.addWeighted(cvs.img, 1 - weight, img2, weight, gamma=0.0)


if __name__ == "__main__":
    img = cv2.imread("insta/samples/Normal.jpg")

    print("Loading image and model")
    F = instafilter("Ludwig")

    print("Applying sampling")
    img2 = F(img)

    cv2.imshow("image", img)
    cv2.waitKey(0)

    cv2.imshow("image", img2)
    cv2.waitKey(0)



from . import Canvas, Artist
from .motion import easing

from contextlib import contextmanager
import cv2
import os
import numpy as np
import scipy.interpolate

from tqdm import tqdm


class Animation:
    def __init__(
        self, width=200, height=200, duration=5, fps=5, extent=4, bg="black"
    ):

        self.width = width
        self.height = height
        self.bg = bg
        self.extent = extent

        self.duration = duration
        self.fps = fps

        n_frames = int(fps * duration)

        self.frames = [
            Canvas(width, height, extent, bg=bg) for _ in range(n_frames)
        ]

        self.has_rendered = [False] * len(self)
        self.timepoints = np.linspace(0, 1, len(self) + 1)[:-1]

        # Keeps track of which list of artists to grab
        self.artist_stack = [[]]
        self.keyframes = np.array([0] * len(self))

    def blank(self, **kwargs):
        """
        Return a new animation object similar to this one. Any arguments passed
        in the keywords are also passed to the new object.
        """
        kws = {
            "width": self.width,
            "height": self.height,
            "bg": self.bg,
            "extent": self.extent,
            "duration": self.duration,
            "fps": self.fps,
        }
        kws.update(kwargs)

        return Animation(**kws)

    def __len__(self):
        return len(self.frames)

    def __call__(self, art):
        self.artist_stack[-1].append(art)
        return self

    def __iadd__(self, rhs):
        """
        Add an Artist to the canvas, or if another animation, extend the
        animation from the rhs.
        """
        if isinstance(rhs, Artist):
            self(rhs)
        elif isinstance(rhs, Animation):
            self.frames.extend(rhs.frames)
            self.has_rendered.extend(rhs.has_rendered)
            self.timepoints = np.hstack([self.timepoints, rhs.timepoints])
            self.duration += rhs.duration
            self.artist_stack.extend(rhs.artist_stack)

            new_keyframes = rhs.keyframes.copy()
            new_keyframes += self.keyframes.max() + 1
            self.keyframes = np.hstack([self.keyframes, new_keyframes])

        else:
            raise TypeError(f"Can't combine a Animation with a {type(rhs)}")

        return self

    def render(self, n):
        assert 0 <= n < len(self)

        if not self.has_rendered[n]:

            t = self.timepoints[n]

            artists = self.artist_stack[self.keyframes[n]]
            for art in artists:
                art(self.frames[n], t)

            self.has_rendered[n] = True

            # Copy any shared attributes to the next frame
            if n < len(self) - 1:
                self.frames[n + 1].shared_attributes.update(
                    self.frames[n].shared_attributes
                )

        return self.frames[n]

    def show(self, delay=50, repeat=True):

        is_status_bar = True
        while True:
            if is_status_bar:
                ITR = tqdm(range(len(self)))
                is_status_bar = False
            else:
                ITR = range(len(self))

            for n in ITR:
                img = self.render(n)
                img.show(delay=delay)
            if not repeat:
                break

    @contextmanager
    def layer(self):
        canvas = _CanvasLayer()
        yield canvas
        self(canvas)


class _CanvasLayer:
    """
    Keep track of the artists when used in a layer.
    """

    def __init__(self):
        self.artists = []

    def __iadd__(self, art):
        self.artists.append(art)
        return self

    def __call__(self, cvs, t=0.0):
        with cvs.layer() as C:
            for art in self.artists:
                art(C, t)

    @contextmanager
    def layer(self):
        msg = "Nested layers in Animation are not possible yet."
        raise NotImplementedError(msg)



import numpy as np
import cv2


def lerp(t, y0, y1, x0=0, x1=1):
    # Linear interpolation
    return np.interp(t, [x0, x1], [y0, y1])


def RGBa_interpolation(t, c0, c1, alpha):
    """
    Interpolate two colors RGB in RGB space parameterized by y. 
    Downweight by alpha.
    """

    R = lerp(t, c0[0], c1[0])
    G = lerp(t, c0[1], c1[1])
    B = lerp(t, c0[2], c1[2])
    a = lerp(t, c0[3], c1[3]) * alpha
    return np.vstack([R, G, B, a]).T


def LABa_interpolation(t, c0, c1, alpha):
    """
    Interpolate two RGB colors in LAB space parameterized by y. 
    Downweight by alpha.
    """

    # Slice off the alpha channel
    a_start, c0 = c0[3], c0[:3]
    a_final, c1 = c1[3], c1[:3]

    # Convert colors to LAB
    C = np.array([c0, c1]).astype(np.uint8).reshape((2, 1, 3))
    c0, c1 = cv2.cvtColor(C, cv2.COLOR_RGB2LAB).reshape(2, 3)

    # Interpolate in LAB space
    L = lerp(t, c0[0], c1[0])
    A = lerp(t, c0[1], c1[1])
    B = lerp(t, c0[2], c1[2])

    # Convert back to RGB
    C = np.vstack([L, A, B]).T.reshape(-1, 1, 3)
    C = C.astype(np.uint8)
    C = cv2.cvtColor(C, cv2.COLOR_LAB2RGB).reshape(-1, 3)

    # Add back in the weighted alpha channel
    a = lerp(t, a_start, a_final)
    a = np.clip(a * alpha, 0, 255).astype(np.uint8)

    # Stack them all together
    C = np.hstack([C, a.reshape(-1, 1)])

    return C



from .artist import Artist
from .canvas import Canvas
from .animation import Animation

from .primitives import circle, line, ellipse, rectangle, polyline, text

from .color import linear_gradient, ColorLoversPalette
from .io import canvas2gif, canvas2mp4

from . import transform
from . import motion
from . import filters
from . import color



"""
Functions to save and load gifs and mp4
"""

from . import Animation
from tqdm import tqdm

import imageio
import tempfile
import os


def canvas2gif(A, f_gif, palettesize=256, gifsicle=False):
    images = [A.render(n).img for n in tqdm(range(len(A)))]

    imageio.mimsave(
        f_gif,
        images,
        duration=A.duration / A.fps,
        palettesize=palettesize,
        subrectangles=True,
    )

    fs = os.stat(f_gif).st_size
    print(f"Rendered {f_gif}, filesize {fs}")

    if gifsicle:
        cmd = f"gifsicle -i {f_gif} --colors {palettesize} -O3 -o {f_gif}"
        os.system(cmd)
        fs = os.stat(f_gif).st_size
        print(f"gifsicle reduced filesize to {fs}")


def canvas2mp4(A, f_mp4, loop=1):

    with tempfile.TemporaryDirectory() as tmp_dir:

        for n, img in tqdm(enumerate(A.frames)):
            A.render(n)
            img.save(f"{tmp_dir}/{n:04d}.png")

        cmd = (
            f"ffmpeg -loop 1 -t {A.duration*loop} "
            f"-y -framerate {A.fps} -i {tmp_dir}/%04d.png "
            f"-c:v libx264 -profile:v high -crf 10 -pix_fmt yuv420p "
            f"{f_mp4}"
        )

        os.system(cmd)

        fs = os.stat(f_mp4).st_size
        print(f"Rendered {f_mp4}, filesize {fs}")



import os
import json
import numpy

_script_path = os.path.dirname(os.path.abspath(__file__))


class NamedColors:
    """
    Named colors from matplotlib
    """

    def __init__(self, f_colors=None):

        if f_colors is None:
            f_colors = os.path.join(
                _script_path, "named_colors", "matplotlib.json"
            )

        with open(f_colors) as FIN:
            raw = FIN.read()

        self.colors = json.loads(raw)

    def __call__(self, name):

        if name not in self.colors:
            color_list = self.colors.keys()
            msg = f"{color_list} are known colors, {name} is unknown."
            raise KeyError(msg)

        color = self.colors[name]

        # Add in the alpha channel
        return color

    def __len__(self):
        return len(self.colors)


if __name__ == "__main__":
    print(P("g"))



import numpy as np
import cv2
import os
from PIL import Image, ImageDraw, ImageFont
from .artist import Artist, constant

_script_path = os.path.dirname(os.path.realpath(__file__))

_DEFAULT_COLOR = "white"
_DEFAULT_SECONDARY_COLOR = "black"
_DEFAULT_THICKNESS = -1
_DEFAULT_ANTIALIASED = True
_DEFAULT_MODE = "direct"
_DEFAULT_FONT_SIZE = 0.5
_DEFAULT_FONT = os.path.join(
    _script_path, "fonts", "Alien-Encounters-Regular.ttf"
)


class PrimitiveArtist(Artist):

    # Basic attributes common to all artists
    x = constant(0.0)
    y = constant(0.0)
    color = constant(_DEFAULT_COLOR)
    thickness = constant(_DEFAULT_THICKNESS)
    antialiased = constant(_DEFAULT_ANTIALIASED)
    mode = constant(_DEFAULT_MODE)
    gradient = constant(None)

    def basic_transforms(self, cvs, t, use_shift=True):
        x = cvs.transform_x(self.x(t), use_shift=use_shift)
        y = cvs.transform_y(self.y(t), use_shift=use_shift)
        thickness = cvs.transform_thickness(self.thickness(t))
        color = cvs.transform_color(self.color(t))
        lineType = cvs.get_lineType(self.antialiased(t))
        mode = self.mode(t)

        # There is a gradient, adjust the mode
        if self.gradient(t) is not None:
            mode = "gradient"

        return x, y, thickness, color, lineType, mode


class circle(PrimitiveArtist):
    """
    Draws a circle to the canvas. Arguments can be a constant 
    or a function that returns a value from t=[0...1].
        
    Args:
    x (float): x coordinate of the circle center
    y (float): y coordinate of the circle center
    r (float): Circle radius

    color (tuple or string): Color (ignored if gradient used)
    thickness (float): Line thickness (use -1 for filled)
    antialiased (bool): Antialiased edges
    gradient (linear_gradient): pixelhouse.gradient for coloring
    mode (string): Direct, blend, add, or subtract
    """

    r = constant(1.0)
    args = ("x", "y", "r", "color", "thickness", "lineType")

    def draw(self, cvs, t=0.0):

        x, y, thickness, color, lineType, mode = self.basic_transforms(cvs, t)
        r = cvs.transform_length(self.r(t), use_shift=True)

        # Require the radius to be non-negative
        r = max(0, r)

        args = (x, y), r, color, thickness, lineType, cvs.shift
        cvs.cv2_draw(cv2.circle, args, mode=mode, gradient=self.gradient, t=t)


class rectangle(PrimitiveArtist):
    """
    Draws a rectangle to the canvas. Arguments can be a constant 
    or a function that returns a value from t=[0...1].
        
    Args:
    x (float): x coordinate of the upper left rectangle
    y (float): y coordinate of the upper left rectangle
    x1 (float): x coordinate of the lower right rectangle
    y1 (float): y coordinate of the lower right rectangle

    color (tuple or string): Color (ignored if gradient used)
    thickness (float): Line thickness (use -1 for filled)
    antialiased (bool): Antialiased edges
    gradient (linear_gradient): pixelhouse.gradient for coloring
    mode (string): Direct, blend, add, or subtract
    """

    x1 = constant(1.0)
    y1 = constant(1.0)
    args = ("x", "y", "x1", "y1", "color", "thickness", "lineType")

    def draw(self, cvs, t=0.0):
        x, y, thickness, color, lineType, mode = self.basic_transforms(cvs, t)
        x1 = cvs.transform_x(self.x1(t), use_shift=True)
        y1 = cvs.transform_y(self.y1(t), use_shift=True)

        args = (x, y), (x1, y1), color, thickness, lineType, cvs.shift
        cvs.cv2_draw(
            cv2.rectangle, args, mode=mode, gradient=self.gradient, t=t
        )


class line(PrimitiveArtist):
    """
    Draws a line to the canvas. Arguments can be a constant 
    or a function that returns a value from t=[0...1].
        
    Args:
    x (float): x coordinate of the upper left point
    y (float): y coordinate of the upper left point
    x1 (float): x coordinate of the lower right point
    y1 (float): y coordinate of the lower right point

    color (tuple or string): Color (ignored if gradient used)
    thickness (float): Line thickness 
    antialiased (bool): Antialiased edges
    gradient (linear_gradient): pixelhouse.gradient for coloring
    mode (string): Direct, blend, add, or subtract
    """

    x1 = constant(1.0)
    y1 = constant(1.0)
    thickness = constant(0.1)
    args = ("x", "y", "x1", "y1", "color", "thickness", "lineType")

    def draw(self, cvs, t=0.0):
        x, y, thickness, color, lineType, mode = self.basic_transforms(cvs, t)
        x1 = cvs.transform_x(self.x1(t), use_shift=True)
        y1 = cvs.transform_y(self.y1(t), use_shift=True)

        # Thickness must be at least one pixel
        thickness = max(thickness, 1)

        args = (x, y), (x1, y1), color, thickness, lineType, cvs.shift
        cvs.cv2_draw(cv2.line, args, mode=mode, gradient=self.gradient, t=t)


class ellipse(PrimitiveArtist):
    """
    Draws an ellipse to the canvas. Arguments can be a constant 
    or a function that returns a value from t=[0...1].
        
    Args:
    x (float): x coordinate of the upper left point
    y (float): y coordinate of the upper left point
    a (float): major axis length 
    b (float): minor axis length
    rotation (float): rotation about the center in radians
    angle_start (float): sector angle to start drawing in radians
    angle_end (float): sector angle to stop drawing in radians

    color (tuple or string): Color (ignored if gradient used)
    thickness (float): Line thickness (use -1 for filled)
    antialiased (bool): Antialiased edges
    gradient (linear_gradient): pixelhouse.gradient for coloring
    mode (string): Direct, blend, add, or subtract
    """

    a = constant(2.0)
    b = constant(1.0)

    rotation = constant(0.0)
    angle_start = constant(0.0)
    angle_end = constant(2 * np.pi)
    args = (
        "x",
        "y",
        "a",
        "b",
        "rotation",
        "angle_start",
        "angle_end",
        "color",
        "thickness",
        "lineType",
    )

    def draw(self, cvs, t=0.0):
        x, y, thickness, color, lineType, mode = self.basic_transforms(cvs, t)

        a = cvs.transform_length(self.a(t), use_shift=True)
        b = cvs.transform_length(self.b(t), use_shift=True)

        rotation = cvs.transform_angle(self.rotation(t))
        angle_start = cvs.transform_angle(self.angle_start(t))
        angle_end = cvs.transform_angle(self.angle_end(t))

        args = (
            (x, y),
            (a, b),
            rotation,
            angle_start,
            angle_end,
            color,
            thickness,
            lineType,
            cvs.shift,
        )

        cvs.cv2_draw(cv2.ellipse, args, mode=mode, gradient=self.gradient, t=t)


class polyline(PrimitiveArtist):
    """
    Draws multiple line segments to the canvas. Arguments can be a constant 
    or a function that returns a value from t=[0...1].
        
    Args:
    xpts (array): x coordinates of the line
    ypts (array): y coordinates of the line
    is_closed(bool): draw the final line segment

    color (tuple or string): Color (ignored if gradient used)
    thickness (float): Line thickness 
    antialiased (bool): Antialiased edges
    gradient (linear_gradient): pixelhouse.gradient for coloring
    mode (string): Direct, blend, add, or subtract
    """

    xpts = constant([0.0, 1.0, 2.0])
    ypts = constant([0.0, 2.0, 0.0])
    thickness = constant(0.1)
    is_closed = constant(1)

    args = ("xpts", "ypts", "is_closed", "color", "thickness", "lineType")

    def draw(self, cvs, t=0.0):

        xpts = cvs.transform_x(np.array(self.xpts(t)), False)
        ypts = cvs.transform_y(np.array(self.ypts(t)), False)

        thickness = cvs.transform_thickness(self.thickness(t))
        color = cvs.transform_color(self.color(t))
        lineType = cvs.get_lineType(self.antialiased(t))
        is_closed = self.is_closed(t)
        mode = self.mode(t)

        pts = np.vstack([xpts, ypts]).T.astype(np.int32)

        if self.gradient(t) is not None:
            raise NotImplementedError("Can't use gradients on polylines yet")

        args = [pts], is_closed, color, thickness, lineType, 0
        cvs.cv2_draw(cv2.polylines, args, mode=mode)


class text(PrimitiveArtist):
    """
    Writes text the canvas. Arguments can be a constant 
    or a function that returns a value from t=[0...1].
        
    Args:
    text (string): Words to draw
    font_size (float): Size of the font (BUG: unknown how this scales)
    font (string): path to font file
    vpos (string): (upper/center/lower) vertical orientation
    hpos (string): (left/center/right) horizontal orientation

    color (tuple or string): Color (ignored if gradient used)
    thickness (float): Line thickness 
    antialiased (bool): Antialiased edges
    gradient (linear_gradient): pixelhouse.gradient for coloring
    mode (string): Direct, blend, add, or subtract
    """

    text = constant("pixelhouse")
    font = constant(_DEFAULT_FONT)
    font_size = constant(_DEFAULT_FONT_SIZE)

    vpos = constant("center")
    hpos = constant("center")

    args = ("text", "fontsize", "vpos", "hpos", "font", "color")

    def draw(self, cvs, t=0.0):
        """
        Use PIL to measure and draw the font.
        """
        x, y, thickness, color, lineType, mode = self.basic_transforms(
            cvs, t, use_shift=False
        )

        text = self.text(t)
        fs = cvs.transform_length(self.font_size(t))
        f_font = self.font(t)

        if not os.path.exists(f_font):
            raise FileNotFoundError(f"Missing font {f_font}")

        font = ImageFont.truetype(self.font(t), fs)

        # Measure the font
        tw, th = font.getsize(text)

        vpos = self.vpos(t)
        if vpos == "upper":
            y -= th
        elif vpos == "center":
            y -= th // 2
        elif vpos == "lower":
            pass
        else:
            possible = ["upper", "center", "lower"]
            raise ValueError(
                f"Unknown font vertical position {vpos}, "
                f"must be one of {possible}."
            )

        hpos = self.hpos(t)
        if hpos == "left":
            pass
        elif hpos == "center":
            x -= tw // 2
        elif hpos == "right":
            x -= tw
        else:
            possible = ["left", "center", "right"]
            raise ValueError(
                f"Unknown font vertical position {hpos}, "
                f"must be one of {possible}."
            )

        # If there is no gradient, just draw it
        if not self.gradient(t):
            pil = Image.fromarray(cvs.copy()._img)

            # Draw the text onto the text canvas
            draw = ImageDraw.Draw(pil)
            draw.text((x, y), text, tuple(color), font)
            cvs._img = np.array(pil)
            return True

        cvs2 = cvs.blank()
        pil = Image.fromarray(cvs2._img)
        draw = ImageDraw.Draw(pil)
        draw.text((x, y), text, (255, 255, 255, 255), font)
        cvs2._img = np.array(pil)

        self.gradient(cvs, t, mask=cvs2)



import numpy as np
import scipy
import itertools


#########################################################################
_DEFAULT_MODE = "blend"


def constant(x):
    def func(self, t=0, *args, **kwargs):
        return x

    return func


class Artist:
    """
    Artists are the backbone of pixelhouse. They draw what's on the screen.
    To be a proper artist, all derived classes must accept their arguments
    as functions and define a "draw" method.
    """

    mode = constant(_DEFAULT_MODE)
    args = []

    @staticmethod
    def _create_interpolation(z):
        t = np.linspace(0, 1, len(z))
        f = scipy.interpolate.interp1d(t, z)

        def func(x):
            return f(x)

        return func

    def __init__(self, *args, **kwargs):
        """
        When an artist is initiated, all of the attributes can be set
        as a function of time. These attributes can be a constant, a numpy
        array (interpolation will be used if needed), or a function.
        """

        for key, val in zip(self.args, args):
            kwargs[key] = val

        attributes = dir(self)
        for key, val in kwargs.items():

            # Can't set attributes an object doesn't have
            if key not in attributes:
                msg = f"{key} not in class {self}"
                raise AttributeError(msg)

            # If the val is callable, that's what we use
            if callable(val):
                setattr(self, key, val)

            # If the val is a numpy array
            elif isinstance(val, np.ndarray):
                interpfunc = self._create_interpolation(val)
                setattr(self, key, interpfunc)

            # Otherwise we assume it's a constant of this value
            else:
                # setattr(self, key, self._constant(val))
                setattr(self, key, constant(val))

    def __call__(self, t, *args, **kwargs):
        # Convenience method for draw
        self.draw(t, *args, **kwargs)

    def draw(self):
        # Virtual class, need to override
        raise NotImplementedError



"""
Base easing functions modified from: https://easings.net/
"""

import math
import numpy as np
from .bezier import bezierMotionCurve


class EasingBase:
    def __init__(self, start=0, stop=1, duration=1.0):
        self.start = start
        self.stop = stop
        self.duration = duration

    def __call__(self):
        t = np.linspace(0, 1, self.duration)
        _, a = self.func(t)
        return self.stop * a + self.start * (1 - a)

    @classmethod
    def func(cls, t):
        raise NotImplementedError


class BezierEase(EasingBase):
    def __init__(
        self, x0=0.45, y0=0.25, x1=0.55, y1=0.75, start=0, stop=1, duration=1
    ):

        super().__init__(start, stop, duration)
        self.f = bezierMotionCurve(x0, y0, x1, y1)

    def func(self, t):
        return self.f(t)


def BezierClassFactory(x0, y0, x1, y1):
    class B(BezierEase):
        def __init__(self, start=0, stop=1, duration=1):
            super().__init__(x0, y0, x1, y1, start, stop, duration)

    return B


named_easings = {
    "easeInSine": (0.47, 0, 0.745, 0.715),
    "easeOutSine": (0.39, 0.575, 0.565, 1),
    "easeInOutSine": (0.445, 0.05, 0.55, 0.95),
    "easeInQuad": (0.55, 0.085, 0.68, 0.53),
    "easeOutQuad": (0.25, 0.46, 0.45, 0.94),
    "easeInOutQuad": (0.455, 0.03, 0.515, 0.955),
    "easeInCubic": (0.55, 0.055, 0.675, 0.19),
    "easeOutCubic": (0.215, 0.61, 0.355, 1),
    "easeInOutCubic": (0.645, 0.045, 0.355, 1),
    "easeInQuart": (0.895, 0.03, 0.685, 0.22),
    "easeOutQuart": (0.165, 0.84, 0.44, 1),
    "easeInOutQuart": (0.77, 0, 0.175, 1),
    "easeInQuint": (0.755, 0.05, 0.855, 0.06),
    "easeOutQuint": (0.23, 1, 0.32, 1),
    "easeInOutQuint": (0.86, 0, 0.07, 1),
    "easeInExpo": (0.95, 0.05, 0.795, 0.035),
    "easeOutExpo": (0.19, 1, 0.22, 1),
    "easeInOutExpo": (1, 0, 0, 1),
    "easeInCirc": (0.6, 0.04, 0.98, 0.335),
    "easeOutCirc": (0.075, 0.82, 0.165, 1),
    "easeInOutCirc": (0.785, 0.135, 0.15, 0.86),
    "easeInBack": (0.6, -0.28, 0.735, 0.045),
    "easeOutBack": (0.175, 0.885, 0.32, 1.275),
    "easeInOutBack": (0.68, -0.55, 0.265, 1.55),
}

for key, vals in named_easings.items():
    locals()[key] = BezierClassFactory(*vals)

#########################################################################
# Custom Easing functions
#########################################################################


class offsetEase(BezierEase):
    def __init__(self, dx=0.0, dy=0.1, baseEase="easeInSine", *args, **kwargs):
        x0, x1, y0, y1 = named_easings[baseEase]
        x0 -= dx
        x1 += dx
        y0 -= dy
        y1 += dy

        super().__init__(x0, x1, y0, y1, *args, **kwargs)


def easeReturn(easing_func, start, stop, frames):
    """
    Returns an easing run that finishes halfway through and returns.
    """
    if isinstance(easing_func, str):
        easing_func = globals()[easing_func]

    n0 = frames // 2
    n1 = frames - n0

    x0 = easing_func(start, stop, n0)()
    x1 = easing_func(stop, start, n1)()

    return np.hstack([x0, x1])



import setuptools
import os

__local__ = os.path.abspath(os.path.dirname(__file__))

f_version = os.path.join(__local__, 'pixelhouse', '_version.py')
exec(open(f_version).read())

# Get the long description from the relevant file
long_description = '''pixelhouse
=================================
A minimalist drawing library for making beautiful animations in python
'''

setuptools.setup(
    name='pixelhouse',
    packages=setuptools.find_packages(),

    # Include package data...
    include_package_data=True,

    description='A minimalist drawing library for making beautiful animations in python',
    long_description=long_description,
    version=__version__,

    # The project's main homepage.
    url="https://github.com/thoppe/pixelhouse",
    #download_url='https://github.com/...',

    # Author details
    author="Travis Hoppe",
    author_email="travis.hoppe+pixelhouse@gmail.com",

    # Choose your license
    license='CC-SA',
    
    install_requires=[
        "opencv-python",
        "Pillow",
        "imageio",
        "numpy",
        "scipy",
        "tqdm",
    ],
    

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5  -Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Education',
        'Intended Audience :: Financial and Insurance Industry',

        'Natural Language :: English',
        
        #'Topic :: Text Processing',
        #'Topic :: Text Processing :: Filters',
        #'Topic :: Utilities',
        
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
    ],

    # What does your project relate to?
    keywords="art",

    test_suite='nose.collector',
    tests_require=['nose'],
)



import os
import json
import numpy

_script_path = os.path.dirname(os.path.abspath(__file__))


class ColorLoversPalette:
    """
    Nice palettes from ...
    """

    def __init__(self, colorset=1000):
        # Load the top 1000 by default

        f_palette = os.path.join(_script_path, f"palettes/{colorset}.json")

        with open(f_palette) as FIN:
            raw = FIN.read()

        self.colors = json.loads(raw)

    def __call__(self, n):
        if n > len(self.colors):
            msg = f"Only {len(self)} palettes known, requested number {n}"
            raise KeyError(msg)

        pal = [x + [255] for x in self.colors[n]]

        return pal

    def __len__(self):
        return len(self.colors)


if __name__ == "__main__":
    P = ColorLoversPalette()
    print(P(2))



from contextlib import contextmanager
import cv2
import numpy as np
import collections
from . import Artist
from .color import matplotlib_colors


class Canvas:
    """
    Basic canvas object for pixelhouse drawings. 
    Extent measures along the x-axis.
    """

    def __init__(
        self,
        width=200,
        height=200,
        extent=4.0,
        bg="black",
        name="pixelhouseImage",
        shift=8,
    ):
        channels = 4
        self._img = np.zeros((height, width, channels), np.uint8)
        self.shift = shift

        # Assign the background color, but make sure it is fully transparent
        # needed for antialiased edges
        self.bg = bg
        bg = np.array(self.transform_color(bg)).astype(np.uint8)
        bg[3] = 0
        self._img[:, :] = bg

        self.name = name
        self.extent = extent

        # When animating, we may want to pass attributes from one canvas
        # to another. Do so in the shared_attributes
        self.shared_attributes = {}

    def __repr__(self):
        return (
            f"pixelhouse (w/h) {self.height}x{self.width}, "
            f"extent {self.extent}"
        )

    @property
    def height(self):
        return self._img.shape[0]

    @property
    def width(self):
        return self._img.shape[1]

    @property
    def channels(self):
        return self._img.shape[2]

    @property
    def img(self):
        return self._img

    @property
    def shape(self):
        return self.height, self.width, self.channels

    @property
    def alpha(self):
        """
        Return the alpha channel from the image
        """
        return self.img[:, :, 3]

    def blank(self, bg=None):
        # Return an empty canvas of the same size
        if bg is None:
            bg = self.bg

        return Canvas(self.width, self.height, bg=bg)

    def copy(self, bg=None, transparent=False):
        # Returns a deep copy of this canvas
        cvs = Canvas(self.width, self.height, bg=self.bg)
        cvs._img = self.img.copy()

        if transparent:
            cvs._img[:, :, 3] = 0

        return cvs

    def __call__(self, art=None):
        """
        Calls an artist on the canvas.
        """
        if art is not None:
            art(self)
        return self

    def __len__(self):
        """
        Return 2 so calling functions work seamlessly between 
        Canvas and Animation (allows for interpolation to not complain).
        """
        return 2

    def combine(self, rhs, mode="blend"):
        if rhs.width != self.width:
            raise ValueError("Can't combine images with different widths")

        if rhs.height != self.height:
            raise ValueError("Can't combine images with different heights")

        if mode == "blend":
            self.blend(rhs)
        elif mode == "add":
            cv2.add(self._img, rhs.img, self._img)
        elif mode == "subtract":
            cv2.subtract(self._img, rhs.img, self._img)
        else:
            raise ValueError(f"Unknown mode {mode}")

        return self

    def __iadd__(self, rhs):
        """
        Add an Artist to the canvas, or combine two canvas depending
        on what the rhs is.
        """
        if isinstance(rhs, Artist):
            self(rhs)
        elif isinstance(rhs, Canvas):
            self.combine(rhs)
        else:
            raise TypeError(f"Can't combine a Canvas with a {type(rhs)}")

        return self

    def cv2_draw(self, func, args, mode, **kwargs):

        if mode == "direct":
            func(self._img, *args)
        elif mode == "gradient":
            rhs = self.blank()
            func(rhs.img, *args)
            kwargs["gradient"](self, t=kwargs["t"], mask=rhs)
        else:
            rhs = self.blank()
            func(rhs.img, *args)
            self.combine(rhs, mode)

    def blend(self, rhs):
        # Smooth the image based off the alpha from the mask image

        a = rhs.alpha.reshape(self.height, self.width, -1).astype(np.float32)
        a /= 255

        MX = (1 - a) * self._img + a * rhs._img
        MX = np.clip(MX, 0, 255).astype(np.uint8)
        self._img = MX

    def transform_x(self, x, is_discrete=True, use_shift=False):
        x *= self.width / 2.0
        x /= self.extent
        x += self.width / 2.0

        if is_discrete:
            if use_shift and self.shift:
                x *= 2 ** self.shift
            return int(x)

        return x

    def inverse_transform_x(self, x):
        x -= self.width / 2
        x *= self.extent
        x /= self.width / 2.0
        return x

    def transform_y(self, y, is_discrete=True, use_shift=False):
        y *= -self.height / 2.0
        y /= self.extent
        y += self.height / 2.0
        if is_discrete:
            if use_shift and self.shift:
                y *= 2 ** self.shift
            return int(y)
        return y

    def inverse_transform_y(self, y):
        y -= self.height / 2
        y *= self.extent
        y /= self.height / 2.0
        return y

    def transform_length(self, r, is_discrete=True, use_shift=False):
        r *= float(self.width) / self.extent

        if use_shift and self.shift:
            r *= 2 ** self.shift

        if is_discrete:
            return int(r)
        return r

    def transform_kernel_length(self, r):
        # Kernels must be positive and odd integers
        r = self.transform_length(r, is_discrete=False)

        remainder = r % 2
        r = int(r - r % 2)
        r += -1 if remainder < 1 else 1

        r = max(1, r)
        return r

    def transform_thickness(self, r):
        # If thickness is negative, leave it alone
        if r > 0:
            return self.transform_length(r)

        return r

    def transform_color(self, c):
        if isinstance(c, str):
            c = matplotlib_colors(c)

        # Force add in the alpha channel
        if len(c) == 3:
            c = list(c) + [255]

        return c

    def grid_coordinates(self):

        attrs = self.shared_attributes

        if (
            "grid_coordinates" in attrs.keys()
            and attrs["prior_shape"] == self.shape
        ):
            return attrs["grid_coordinates"]

        shape = tuple(self.shape)

        xg, yg, zg = np.meshgrid(
            np.arange(shape[1]), np.arange(shape[0]), np.arange(shape[2])
        )

        attrs["grid_coordinates"] = (xg, yg, zg)
        attrs["prior_shape"] = self.shape

        return attrs["grid_coordinates"]

    def grid_points(self):

        attrs = self.shared_attributes

        if "grid_points" in attrs.keys() and attrs["prior_shape"] == self.shape:
            return attrs["grid_points"]

        xg, yg, _ = self.grid_coordinates()
        xg = xg[:, :, 0]
        yg = yg[:, :, 0]

        xp = self.inverse_transform_x(xg.astype(np.float32))
        yp = self.inverse_transform_y(yg.astype(np.float32))

        attrs["grid_points"] = (xp, yp)
        attrs["prior_shape"] = self.shape

        return attrs["grid_points"]

    @staticmethod
    def transform_angle(rads):
        # From radians into degrees, counterclockwise
        return -rads * (360 / (2 * np.pi))

    @staticmethod
    def get_lineType(antialiased):
        if antialiased:
            return cv2.LINE_AA
        return 8

    def show(self, delay=0):
        # Before we show we have to convert back to BGR
        dst = cv2.cvtColor(self.img, cv2.COLOR_RGB2BGR)

        cv2.imshow(self.name, dst)
        cv2.waitKey(delay)

    def save(self, f_save):
        # Before we save we have to convert back to BGR
        dst = cv2.cvtColor(self.img, cv2.COLOR_RGB2BGR)
        cv2.imwrite(f_save, dst)

    def load(self, f_img):
        # Read the image in and convert to RGB space
        self._img = cv2.cvtColor(cv2.imread(f_img), cv2.COLOR_BGR2RGB)

        alpha = np.zeros_like(self._img[:, :, 0])
        self._img = np.dstack((self._img, alpha))

        return self

    @contextmanager
    def layer(self):
        canvas = self.copy(transparent=True)
        yield canvas
        self += canvas



from .easing import easeInSine, easeOutSine, easeInOutSine
from .easing import easeInQuad, easeOutQuad, easeInOutQuad
from .easing import easeInCubic, easeOutCubic, easeInOutCubic
from .easing import easeInQuart, easeOutQuart, easeInOutQuart
from .easing import easeInQuint, easeOutQuint, easeInOutQuint
from .easing import easeInExpo, easeOutExpo, easeInOutExpo
from .easing import easeInCirc, easeOutCirc, easeInOutCirc
from .easing import easeInBack, easeOutBack, easeInOutBack

from .easing import offsetEase, easeReturn



import numpy as np
import cv2
from ..artist import Artist, constant


class scale(Artist):
    fx = constant(0.5)
    fy = constant(None)
    args = ("fx", "fy")

    def draw(self, cvs, t=0.0):

        fx, fy = self.fx(t), self.fy(t)
        if fy is None:
            fy = fx

        cvs._img = cv2.resize(cvs.img, (0, 0), fx=fx, fy=fy)


class translate(Artist):
    x = constant(0.5)
    y = constant(0.0)
    args = ("x", "y")

    def draw(self, cvs, t=0.0):
        x = cvs.transform_length(self.x(t))
        y = cvs.transform_length(self.y(t))
        bg = cvs.transform_color(cvs.bg)

        M = np.float32([[1, 0, x], [0, 1, y]])
        cvs._img = cv2.warpAffine(cvs.img, M, cvs.shape[:2], borderValue=bg)


class rotate(Artist):
    theta = constant(np.pi / 4)
    x = constant(0.0)
    y = constant(0.0)
    args = ("theta", "x", "y")

    def draw(self, cvs, t=0.0):
        theta = cvs.transform_angle(self.theta(t))
        x = cvs.transform_x(self.x(t))
        y = cvs.transform_y(self.y(t))
        bg = cvs.transform_color(cvs.bg)

        cols, rows = cvs.shape[:2]
        M = cv2.getRotationMatrix2D((x, y), theta, 1)
        cvs._img = cv2.warpAffine(cvs.img, M, (cols, rows), borderValue=bg)
